/*
 *
 * VMMZ: Delta Encoding of Virtual-Machine Memory
 * Copyright (C) 2015-2016  James E. Fowler
 *
 * This file is a part of VMMZ.
 *
 * VMMZ is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with VMMZ; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


#include "vmmz.h"


int VMMZGetFileSize(const QccString filename, long *filesize)
{
  int return_value = 0;
  FILE *infile = NULL;
  
  if (!strcmp(&filename[strlen(filename) - 3], ".gz"))
    {
      *filesize = 0;
      if ((infile = QccFileOpen(filename, "r")) == NULL)
        {
          QccErrorAddMessage("(VMMZGetFileSize): Error calling QccFileOpen()");
          goto Error;
        }
      while (fgetc(infile) != EOF)
        (*filesize)++;
    }
  else
    if (QccFileGetSize(filename, NULL, filesize))
      {
        QccErrorAddMessage("(VMMZGetFileSize): Error calling QccFileGetSize()");
        goto Error;
      }
  
  return_value = 0;
  goto Return;
 Error:
  return_value = 1;
 Return:
  QccFileClose(infile);
  return(return_value);
}


static unsigned char *VMMZReadFile(const QccString filename,
                                   long *filesize)
{
  FILE *infile = NULL;
  long filesize2 = 0;
  unsigned char *mem_buffer = NULL;
  long index = 0;
  
  if (VMMZGetFileSize(filename, &filesize2))
    {
      QccErrorAddMessage("(VMMZEncode): Error calling VMMZGetFileSize()");
      goto Error;
    }
  
  if ((infile = QccFileOpen(filename, "r")) == NULL)
    {
      QccErrorAddMessage("(VMMZReadFile): Error calling QccFileOpen()");
      goto Error;
    }
  
  if ((mem_buffer =
       (unsigned char *)malloc(sizeof(unsigned char) * filesize2)) == NULL)
    {
      QccErrorAddMessage("(VMMZReadFile): Error allocating memory");
      goto Error;
    }
  
  for (index = 0; index < filesize2; index++)
    if (QccFileReadChar(infile, (char *)(&mem_buffer[index])))
      {
        QccErrorAddMessage("(VMMZReadFile): Error calling QccFileReadChar()");
        goto Error;
      }
  
  goto Return;
 Error:
  if (mem_buffer != NULL)
    free(mem_buffer);
  mem_buffer = NULL;
 Return:
  QccFileClose(infile);
  if (filesize != NULL)
    *filesize = filesize2;
  return(mem_buffer);
}


void VMMZSetHeader(void)
{
  QccString header;
  
  QccStringSprintf(header,
                   "-------------------------------------------------------------------------------\n\n    %s Version %d.%d %s\n    %s\n\n-------------------------------------------------------------------------------\n\nemploying:",
                   "VMMZ",
                   VMMZ_MAJORVERSION,
                   VMMZ_MINORVERSION,
                   VMMZ_DATE,
                   VMMZ_COPYRIGHT);
  
  QccSetUserHeader(header);
}


int VMMZDiff(const QccString filename1,
             const QccString filename2,
             long *filesize,
             long *num_differences,
             QccIMGImage *output_image)
{
  FILE *infile1 = NULL;
  FILE *infile2 = NULL;
  long filesize1 = 0;
  long filesize2 = 0;
  long index = 0;
  long cnt = 0;
  char ch1;
  char ch2;
  int return_value = 0;
  int num_rows = 0;
  int num_cols = 0;
  int row = 0;
  int col = 0;
  
  if (VMMZGetFileSize(filename1, &filesize1))
    {
      QccErrorAddMessage("(VMMZDiff): Error calling VMMZGetFileSize()");
      goto Error;
    }
  
  if (VMMZGetFileSize(filename2, &filesize2))
    {
      QccErrorAddMessage("(VMMZDiff): Error calling VMMZGetFileSize()");
      goto Error;
    }
  
  if (filesize1 != filesize2)
    {
      QccErrorAddMessage("(VMMZEncode): files are not the same size");
      goto Error;
    }
  
  if ((infile1 = QccFileOpen(filename1, "r")) == NULL)
    {
      QccErrorAddMessage("(VMMZDiff): Error calling QccFileOpen()");
      goto Error;
    }
  
  if ((infile2 = QccFileOpen(filename2, "r")) == NULL)
    {
      QccErrorAddMessage("(VMMZDiff): Error calling QccFileOpen()");
      goto Error;
    }
  
  if (output_image != NULL)
    {
      for (num_rows = ((int)floor(sqrt(filesize1)));
           filesize1 % num_rows;
           num_rows--)
        ;
      num_cols = filesize1 / num_rows;
      
      output_image->image_type = QCCIMGTYPE_PGM;
      if (QccIMGImageSetSize(output_image,
                             num_rows,
                             num_cols))
        {
          QccErrorAddMessage("(VMMZDiff): Error calling QccIMGImageSetSize()");
          goto Error;
        }
      if (QccIMGImageAlloc(output_image))
        {
          QccErrorAddMessage("(VMMZDiff): Error calling QccIMGImageAlloc()");
          goto Error;
        }
    }
  
  for (cnt = 0, index = 0, row = 0, col = 0;
       index < filesize1; index++)
    {
      if (QccFileReadChar(infile1, &ch1))
        {
          QccErrorAddMessage("(VMMZDiff): Error calling QccFileReadChar()");
          goto Error;
        }
      if (QccFileReadChar(infile2, &ch2))
        {
          QccErrorAddMessage("(VMMZDiff): Error calling QccFileReadChar()");
          goto Error;
        }
      
      if (ch1 == ch2)
        {
          if (output_image != NULL)
            output_image->Y.image[row][col] = 255;
        }
      else
        {
          if (output_image != NULL)
            output_image->Y.image[row][col] = 0;
          cnt++;
        }
      
      if (output_image != NULL)
        {
          col++;
          if (col >= num_cols)
            {
              row++;
              col = 0;
            }
        }
    }
  
  return_value = 0;
  goto Return;
 Error:
  return_value = 1;
 Return:
  QccFileClose(infile1);
  QccFileClose(infile2);
  if (num_differences != NULL)
    *num_differences = cnt;
  if (filesize != NULL)
    *filesize = filesize1;
  return(return_value);
}


static int VMMZEncodeHeader(FILE *outfile,
                            const QccString baseline_filename,
                            const QccString vmdk_filename,
                            const VMMZCopyTable *copy_table)
{
  int return_value = 0;
  QccString baseline_filepath;
  QccString vmdk_filepath;
  
  if (QccFileGetRealPath(baseline_filename,
                         baseline_filepath))
    {
      QccErrorAddMessage("(VMMZEncodeheader): Error calling QccFileGetRealPath()");
      goto Error;
    }
  
  if (QccFileGetRealPath(vmdk_filename,
                         vmdk_filepath))
    {
      QccErrorAddMessage("(VMMZEncodeheader): Error calling QccFileGetRealPath()");
      goto Error;
    }
  
  if (QccFileWriteMagicNumberVersion(outfile,
                                     VMMZ_MAGICNUM,
                                     VMMZ_MAJORVERSION,
                                     VMMZ_MINORVERSION))
    {
      QccErrorAddMessage("(VMMZEncodeHeader): Error calling QccFileWriteMagicNumberVersion()");
      goto Error;
    }
  
  if (QccFileWriteString(outfile,
                         "baseline:"))
    {
      QccErrorAddMessage("(VMMZEncodeHeader): Error calling QccFileWriteString()");
      goto Error;
    }
  if (QccFileWriteString(outfile,
                         baseline_filepath))
    {
      QccErrorAddMessage("(VMMZEncodeHeader): Error calling QccFileWriteString()");
      goto Error;
    }
  
  if (QccFileWriteChar(outfile,
                       '\n'))
    {
      QccErrorAddMessage("(VMMZEncodeHeader): Error calling QccFileWriteChar()");
      goto Error;
    }
  
  if (QccFileWriteString(outfile,
                         "vmdk:"))
    {
      QccErrorAddMessage("(VMMZEncodeHeader): Error calling QccFileWriteString()");
      goto Error;
    }
  if (QccFileWriteString(outfile,
                         vmdk_filepath))
    {
      QccErrorAddMessage("(VMMZEncodeHeader): Error calling QccFileWriteString()");
      goto Error;
    }
  
  if (QccFileWriteChar(outfile,
                       '\n'))
    {
      QccErrorAddMessage("(VMMZEncodeHeader): Error calling QccFileWriteChar()");
      goto Error;
    }

  if (VMMZWriteCopyTable(outfile, copy_table))
    {
      QccErrorAddMessage("(VMMZEncodeHeader): Error calling VMMZWriteCopyTable())");
      goto Error;
    }
    
  return_value = 0;
  goto Return;
 Error:
  return_value = 1;
 Return:
  return(return_value);
}


static int VMMZDecodeHeader(FILE *infile,
                            QccString baseline_filename,
                            QccString vmdk_filename,
                            long *header_size,
                            VMMZCopyTable *copy_table)
{
  int return_value = 0;
  QccString magic_number;
  QccString baseline_filename2;
  QccString vmdk_filename2;
  
  if (QccFileReadMagicNumber(infile,
                             magic_number,
                             NULL,
                             NULL))
    {
      QccErrorAddMessage("(VMMZDecodeHeader): Error calling QccFileReadMagicNumber()");
      goto Error;
    }
  
  if (strcmp(magic_number,
             VMMZ_MAGICNUM))
    {
      QccErrorAddMessage("(VMMZDecodeHeader): infile is not of compressed VMMZ (%s) type",
                         VMMZ_MAGICNUM);
      goto Error;
    }
  
  if (QccFileReadLine(infile, baseline_filename2))
    {
      QccErrorAddMessage("(VMMZDecodeHeader): Error calling QccFileReadLine()");
      goto Error;
    }
  
  if (strstr(baseline_filename2, "baseline:") != baseline_filename2)
    {
      QccErrorAddMessage("(VMMZDecodeHeader): Error reading baseline filename");
      goto Error;
    }
  
  QccStringCopy(baseline_filename, &baseline_filename2[9]);
  
  if (QccFileReadLine(infile, vmdk_filename2))
    {
      QccErrorAddMessage("(VMMZDecodeHeader): Error calling QccFileReadLine()");
      goto Error;
    }
  
  if (strstr(vmdk_filename2, "vmdk:") != vmdk_filename2)
    {
      QccErrorAddMessage("(VMMZDecodeHeader): Error reading VMDK filename");
      goto Error;
    }
  
  QccStringCopy(vmdk_filename, &vmdk_filename2[5]);

  if (QccFileSkipWhiteSpace(infile, 1))
    {
      QccErrorAddMessage("(VMMZDecodeHeader): Error calling QccFileSkipWhiteSpace()");
      goto Error;
    }
  
  if (VMMZReadCopyTable(infile, copy_table))
    {
      QccErrorAddMessage("(VMMZDecodeHeader): Error calling VMMZReadCopyTable()");
      goto Error;
    }
    
  if (header_size != NULL)
    if (QccFileGetCurrentPosition(infile, header_size))
      {
        QccErrorAddMessage("(VMMZDecodeHeader): Error calling QccFileGetCurrentPosition()");
        goto Error;
      }
  
  return_value = 0;
  goto Return;
 Error:
  return_value = 1;
 Return:
  return(return_value);
}


int VMMZEncode(const QccString baseline_filename,
               const QccString current_vmem_filename,
               const QccString outfile_filename,
               const QccString vmdk_filename,
               const QccString volatility_profile,
               long *baseline_filesize,
               const QccString predicted_vmem_filename)
{
  int return_value = 0;
  long baseline_filesize2 = 0;
  long current_vmem_filesize = 0;
  FILE *outfile = NULL;
  unsigned char *baseline_mem = NULL;
  unsigned char *current_vmem = NULL;
  unsigned char *output_mem = NULL;
  usize_t output_size = 0;
  usize_t available_output_size = 0;
  VMMZPETable pe_table;
  VMMZProcessTable process_table;
  VMMZCopyTable copy_table;
  
  VMMZPETableInitialize(&pe_table);
  VMMZProcessTableInitialize(&process_table);
  VMMZCopyTableInitialize(&copy_table);
  
  if (!QccFileExists(baseline_filename))
    {
      QccErrorAddMessage("(VMMZEncode): file %s does not exist",
                         baseline_filename);
      goto Error;
    }
  if (!QccFileExists(current_vmem_filename))
    {
      QccErrorAddMessage("(VMMZEncode): file does not exist");
      goto Error;
    }
  if (!QccFileExists(vmdk_filename))
    {
      QccErrorAddMessage("(VMMZEncode): file %s does not exist",
                         vmdk_filename);
      goto Error;
    }
  
  if ((outfile = QccFileOpen(outfile_filename,
                             "w")) == NULL)
    {
      QccErrorAddMessage("(VMMZEncode): Error calling QccFileOpen()");
      goto Error;
    }
  
  //  printf("Getting PE info...\n");
  if (VMMZGetPETable(baseline_filename,
                     current_vmem_filename,
                     vmdk_filename,
                     volatility_profile,
                     &pe_table,
                     &process_table,
                     &copy_table))
    {
      QccErrorAddMessage("(VMMZEncode): Error calling VMMZGetPETable()");
      goto Error;
    }
  
  //  printf("Reading baseline...\n");
  if ((baseline_mem = VMMZReadFile(baseline_filename,
                                   &baseline_filesize2)) == NULL)
    {
      QccErrorAddMessage("(VMMZEncode): Error calling VMMZReadFile()");
      goto Error;
    }
  
  //  printf("Inserting PEs...\n");
  if (VMMZInsertPEs(baseline_mem,
                    baseline_filesize2,
                    &pe_table,
                    &process_table,
                    &copy_table,
                    predicted_vmem_filename))
    {
      QccErrorAddMessage("(VMMZEncode): Error calling VMMZInsertPEs()");
      goto Error;
    }
  
  //  printf("Writing header...\n");
  if (VMMZEncodeHeader(outfile,
                       baseline_filename,
                       vmdk_filename,
                       &copy_table))
    {
      QccErrorAddMessage("(VMMZEncode): Error calling VMMZEncodeHeader()");
      goto Error;
    }
  
  //  printf("Reading infile...\n");
  if ((current_vmem = VMMZReadFile(current_vmem_filename,
                                   &current_vmem_filesize)) == NULL)
    {
      QccErrorAddMessage("(VMMZEncode): Error calling VMMZReadFile()");
      goto Error;
    }
  
  if (baseline_filesize2 != current_vmem_filesize)
    {
      QccErrorAddMessage("(VMMZEncode): baseline and infile are not the same size");
      goto Error;
    }
  
  available_output_size = floor(VMMZ_ENCODED_OUTFILESIZE * current_vmem_filesize);
  
  if ((output_mem =
       (unsigned char *)malloc(sizeof(unsigned char) *
                               available_output_size)) ==
      NULL)
    {
      QccErrorAddMessage("(VMMZEncode): Error allocating memory");
      goto Error;
    }
  
  //  printf("xdelta3 encoding...\n");
  
  if ((return_value = xd3_encode_memory((uint8_t *)current_vmem,
                                        (usize_t)current_vmem_filesize,
                                        (uint8_t *)baseline_mem,
                                        (usize_t)current_vmem_filesize,
                                        (uint8_t *)output_mem,
                                        &output_size,
                                        available_output_size,
                                        XD3_SEC_DJW | XD3_COMPLEVEL_9)))
    {
      QccErrorAddMessage("(VMMZEncode): Error calling xd3_encode_memory() - return: %d", return_value);
      goto Error;
    }
  
  //  printf("Writing outfile... %d\n", output_size);
  
  if (fwrite((void *)output_mem,
             sizeof(unsigned char),
             (size_t)output_size,
             outfile) != output_size)
    {
      QccErrorAddMessage("(VMMZEncode): Error writing output file");
      goto Error;
    }
  
  return_value = 0;
  goto Return;
 Error:
  return_value = 1;
 Return:
  QccFileClose(outfile);
  VMMZUnmountVMDK(pe_table.mount_point);
  VMMZPETableFree(&pe_table);
  VMMZProcessTableFree(&process_table);
  VMMZCopyTableFree(&copy_table);
  if (output_mem != NULL)
    free(output_mem);
  if (baseline_mem != NULL)
    free(baseline_mem);
  if (current_vmem != NULL)
    free(current_vmem);
  if (baseline_filesize != NULL)
    *baseline_filesize = baseline_filesize2;
  return(return_value);
}


int VMMZDecode(const QccString baseline_filename,
               const QccString vmdk_filename,
               const QccString infile_filename,
               const QccString output_vmem_filename)
{
  int return_value = 0;
  FILE *outfile = NULL;
  FILE *infile = NULL;
  QccString baseline_filename2;
  QccString vmdk_filename2;
  unsigned char *baseline_mem = NULL;
  unsigned char *output_mem = NULL;
  unsigned char *infile_mem = NULL;
  long baseline_filesize2 = 0;
  long header_size = 0;
  long infile_size = 0;
  usize_t output_size = 0;
  VMMZCopyTable copy_table;
  QccString mount_point;
  
  VMMZCopyTableInitialize(&copy_table);

  if ((infile = QccFileOpen(infile_filename,
                            "r")) == NULL)
    {
      QccErrorAddMessage("(VMMZDecode): Error calling QccFileOpen()");
      goto Error;
    }
  
  //  printf("Reading header...\n");
  if (VMMZDecodeHeader(infile,
                       baseline_filename2,
                       vmdk_filename2,
                       &header_size,
                       &copy_table))
    {
      QccErrorAddMessage("(VMMZDecode): Error calling VMMZDecodeHeader()");
      goto Error;
    }
  
  //  printf("header size = %ld\n", header_size);
  
  if (baseline_filename != NULL)
    QccStringCopy(baseline_filename2, baseline_filename);
  
  if (vmdk_filename != NULL)
    QccStringCopy(vmdk_filename2, vmdk_filename);
    
  //  printf("Reading baseline %s...\n", baseline_filename2);
  if ((baseline_mem = VMMZReadFile(baseline_filename2,
                                   &baseline_filesize2)) == NULL)
    {
      QccErrorAddMessage("(VMMZDecode): Error calling VMMZReadFile()");
      goto Error;
    }
  
  QccConvertToQccString(mount_point, VMMZ_TMPDIR);

  if (VMMZMountVMDK(vmdk_filename2,
                    mount_point))
    {
      QccErrorAddMessage("(VMMZDecode): Error calling VMMZMountVMDK()");
      goto Error;
    }

  //  printf("Inserting copy table...\n");
  if (VMMZCopyTableInsert(baseline_mem,
                          baseline_filesize2,
                          &copy_table,
                          mount_point))
    {
      QccErrorAddMessage("(VMMZDecode): Error calling VMMZCopyTableInsert()");
      goto Error;
    }
  
  //  printf("Reading infile...\n");
  if (VMMZGetFileSize(infile_filename,
                      &infile_size))
    {
      QccErrorAddMessage("(VMMZDecode): Error calling VMMZGetFileSize()");
      goto Error;
    }
  
  infile_size -= header_size;
  
  //  printf("infile size = %ld\n", infile_size);
  
  if ((infile_mem =
       (unsigned char*)malloc(sizeof(unsigned char) * infile_size)) == NULL)
    {
      QccErrorAddMessage("(VMMZDecode): Error allocating memory");
      goto Error;
    }
  
  if ((fread((void *)infile_mem,
             sizeof(unsigned char),
             (size_t)infile_size,
             infile)) != infile_size)
    {
      QccErrorAddMessage("(VMMZDecode): Error reading input file %s",
                         infile_filename);
      goto Error;
    }
  
  if ((output_mem =
       (unsigned char *)malloc(sizeof(unsigned char) * baseline_filesize2)) ==
      NULL)
    {
      QccErrorAddMessage("(VMMZDecode): Error allocating memory");
      goto Error;
    }
  
  //  printf("xd3 decoding...\n");
  if ((return_value = xd3_decode_memory((uint8_t *)infile_mem,
                                        (usize_t)infile_size,
                                        (uint8_t *)baseline_mem,
                                        (usize_t)baseline_filesize2,
                                        (uint8_t *)output_mem,
                                        &output_size,
                                        (usize_t)baseline_filesize2,
                                        XD3_SEC_DJW | XD3_COMPLEVEL_9)))
    {
      QccErrorAddMessage("(VMMZDecode): Error calling xd3_decode_memory() - return: %d", return_value);
      goto Error;
    }
  
  if (output_size != baseline_filesize2)
    {
      QccErrorAddMessage("(VMMZDecode): Output file is not same size as baseline file");
      goto Error;
    }
  
  //  printf("Writing outfile...\n");
  
  if ((outfile = QccFileOpen(output_vmem_filename, "w")) == NULL)
    {
      QccErrorAddMessage("(VMMZDecode): Error calling QccFileOpen()");
      goto Error;
    }
  
  if (fwrite((void *)output_mem,
             sizeof(unsigned char),
             baseline_filesize2,
             outfile) != baseline_filesize2)
    {
      QccErrorAddMessage("(VMMZDecode): Error writing output file");
      goto Error;
    }
  
  return_value = 0;
  goto Return;
 Error:
  return_value = 1;
 Return:
  QccFileClose(infile);
  QccFileClose(outfile);
  VMMZUnmountVMDK(mount_point);
  VMMZCopyTableFree(&copy_table);
  if (infile_mem != NULL)
    free(infile_mem);
  if (output_mem != NULL)
    free(output_mem);
  if (baseline_mem != NULL)
    free(baseline_mem);
  return(return_value);
}
