/*
 *
 * VMMZ: Delta Encoding of Virtual-Machine Memory
 * Copyright (C) 2015-2016  James E. Fowler
 *
 * This file is a part of VMMZ.
 *
 * VMMZ is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with VMMZ; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


#include "vmmz.h"


void VMMZCopyTableInitialize(VMMZCopyTable *copy_table)
{
  if (copy_table != NULL)
    {
      copy_table->num_files = 0;
      copy_table->files = NULL;
    }
}


int VMMZCopyTableAlloc(VMMZCopyTable *copy_table)
{
  int index;
  
  if (copy_table != NULL)
    if (copy_table->num_files > 0)
      {
        if ((copy_table->files =
             (VMMZCopyFile *)malloc(sizeof(VMMZCopyFile) *
                                    copy_table->num_files)) == NULL)
          {
            QccErrorAddMessage("(VMMZCopyTableAlloc): Error allocating memory");
            return(1);
          }
        for (index = 0; index < copy_table->num_files; index++)
          VMMZCopyFileInitialize(&copy_table->files[index]);
      }

  return(0);
}


void VMMZCopyTableFree(VMMZCopyTable *copy_table)
{
  int index;
  
  if (copy_table != NULL)
    if (copy_table->files != NULL)
      {
        for (index = 0; index < copy_table->num_files; index++)
          VMMZCopyFileFree(&copy_table->files[index]);
        free(copy_table->files);
      }
}


int VMMZWriteCopyTable(FILE *outfile,
                       const VMMZCopyTable *copy_table)
{
  int return_value = 0;
  int file_cnt;
  int copy_cnt;
  
  if (outfile == NULL)
    return(0);
  if (copy_table == NULL)
    return(0);
  
  if (QccFileWriteInt(outfile,
                      copy_table->num_files))
    {
      QccErrorAddMessage("(VMMZWriteCopyTable): Error calling QccFileWriteInt()");
      goto Error;
    }
  
  for (file_cnt = 0; file_cnt < copy_table->num_files; file_cnt++)
    {
      if (QccFileWriteString(outfile, copy_table->files[file_cnt].filename))
        {
          QccErrorAddMessage("(VMMZWriteCopyTable): Error calling QccFileWriteString()");
          goto Error;
        }
      if (QccFileWriteChar(outfile, '\n'))
        {
          QccErrorAddMessage("(VMMZWriteCopyTable): Error calling QccFileWriteChar()");
          goto Error;
        }
      if (QccFileWriteInt(outfile,
                          copy_table->files[file_cnt].num_copies))
        {
          QccErrorAddMessage("(VMMZWriteCopyTable): Error calling QccFileWriteInt()");
          goto Error;
        }
      for (copy_cnt = 0; copy_cnt < copy_table->files[file_cnt].num_copies;
           copy_cnt++)
        {
          if (QccFileWriteInt(outfile,
                              copy_table->files[file_cnt].copies[copy_cnt].source))
            {
              QccErrorAddMessage("(VMMZWriteCopyTable): Error calling QccFileWriteInt()");
              goto Error;
            }
          if (QccFileWriteInt(outfile,
                              copy_table->files[file_cnt].copies[copy_cnt].destination))
            {
              QccErrorAddMessage("(VMMZWriteCopyTable): Error calling QccFileWriteInt()");
              goto Error;
            }
          if (QccFileWriteInt(outfile,
                              copy_table->files[file_cnt].copies[copy_cnt].length))
            {
              QccErrorAddMessage("(VMMZWriteCopyTable): Error calling QccFileWriteInt()");
              goto Error;
            }
        }
    }
  
  return_value = 0;
  goto Return;
 Error:
  return_value = 1;
 Return:
  return(return_value);
}


int VMMZReadCopyTable(FILE *infile,
                      VMMZCopyTable *copy_table)
{
  int return_value = 0;
  int file_cnt;
  int copy_cnt;
  int num_copies;
  VMMZCopyEntry copy;
  
  if (infile == NULL)
    return(0);
  if (copy_table == NULL)
    return(0);
  
  if (QccFileReadInt(infile,
                     &copy_table->num_files))
    {
      QccErrorAddMessage("(VMMZReadCopyTable): Error calling QccFileReadInt()");
      goto Error;
    }

  //  printf("%d\n", copy_table->num_files);
  
  if (VMMZCopyTableAlloc(copy_table))
    {
      QccErrorAddMessage("(VMMZReadCopyTable): Error calling VMMZCopyTableAlloc()");
      goto Error;
    }
    
  for (file_cnt = 0; file_cnt < copy_table->num_files; file_cnt++)
    {
      if (QccFileReadLine(infile, copy_table->files[file_cnt].filename))
        {
          QccErrorAddMessage("(VMMZReadCopyTable): Error calling QccFileReadString()");
          goto Error;
        }

      if (QccFileReadInt(infile, &num_copies))
        {
          QccErrorAddMessage("(VMMZReadCopyTable): Error calling QccFileReadInt()");
          goto Error;
        }
      for (copy_cnt = 0; copy_cnt < num_copies; copy_cnt++)
        {
          if (QccFileReadInt(infile,
                             (int *)(&copy.source)))
            {
              QccErrorAddMessage("(VMMZReadCopyTable): Error calling QccFileReadInt()");
              goto Error;
            }
          if (QccFileReadInt(infile,
                             (int *)(&copy.destination)))
            {
              QccErrorAddMessage("(VMMZReadCopyTable): Error calling QccFileReadInt()");
              goto Error;
            }
          if (QccFileReadInt(infile,
                             (int *)(&copy.length)))
            {
              QccErrorAddMessage("(VMMZReadCopyTable): Error calling QccFileReadInt()");
              goto Error;
            }
          if (VMMZCopyFileAddCopy(&copy_table->files[file_cnt], &copy))
            {
              QccErrorAddMessage("(VMMZReadCopyTable): Error calling VMMZCopyFileAddCopy()");
              goto Error;
            }
        }
    }
  
  return_value = 0;
  goto Return;
 Error:
  return_value = 1;
 Return:
  return(return_value);
}


int VMMZCopyTableInsert(unsigned char *memory,
                        long memory_size,
                        VMMZCopyTable *copy_table,
                        const QccString mount_point)
{
  int return_value = 0;
  int file;
  
  if (memory == NULL)
    return(0);
  if (copy_table == NULL)
    return(0);

  for (file = 0; file < copy_table->num_files; file++)
    if (VMMZCopyFileInsert(memory,
                           memory_size,
                           &copy_table->files[file],
                           mount_point))
      {
        QccErrorAddMessage("(VMMZCopyTableInsert): Error calling VMMZCopyFileInsert()");
        goto Error;
      }
  
  return_value = 0;
  goto Return;
 Error:
  return_value = 1;
 Return:
  return(return_value);
}
