/*
 *
 * VMMZ: Delta Encoding of Virtual-Machine Memory
 * Copyright (C) 2015-2016  James E. Fowler
 *
 * This file is a part of VMMZ.
 *
 * VMMZ is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with VMMZ; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


#include "vmmz.h"


int VMMZGetPETable(const QccString baseline_filename,
                   const QccString current_vmem_filename,
                   const QccString vmdk_filename,
                   const QccString volatility_profile,
                   VMMZPETable *pe_table,
                   VMMZProcessTable *process_table,
                   VMMZCopyTable *copy_table)

{
  int return_value = 0;
  FILE *infile = NULL;
  QccString cmd;
  int num_exes = 0;
  int num_dlls = 0;
  int num_processes = 0;
  int pe_cnt = 0;
  int process_cnt = 0;
  int page_cnt = 0;
  
  QccConvertToQccString(pe_table->mount_point, VMMZ_TMPDIR);
  
  if (VMMZMountVMDK(vmdk_filename,
                    pe_table->mount_point))
    {
      QccErrorAddMessage("(VMMZGetPETable): Error calling VMMZMountVMDK()");
      goto Error;
    }
  
  QccStringSprintf(cmd,
                   "%s/vmmz_parse %s %s %s %s %s %s",
                   VMMZ_INSTALL_PATH_STR,
                   VMMZ_PARSE_MAGIC_NUM,
                   VMMZ_VOLATILITY_CMD_STR,
                   volatility_profile,
                   pe_table->mount_point,
                   baseline_filename,
                   current_vmem_filename);
  
  if ((infile = popen(cmd, "r")) == NULL)
    {
      QccErrorAddMessage("(VMMZGetPETable): Error opening pipe to vmmz_parse");
      goto Error;
    }
  
  if (fscanf(infile, "%d,%d,%d\n",
             &num_exes, &num_dlls, &num_processes) != 3)
    {
      QccErrorAddMessage("(VMMZGetPETable): Error reading from vmmz_parse");
      goto Error;
    }
  
  pe_table->num_pes = num_exes + num_dlls;
  if (VMMZPETableAlloc(pe_table))
    {
      QccErrorAddMessage("(VMMZGetPETable): Error calling VMMZPETableAlloc()");
      goto Error;
    }
    
  process_table->num_processes = num_processes;
  if (VMMZProcessTableAlloc(process_table))
    {
      QccErrorAddMessage("(VMMZGetPETable): Error calling VMMZProcessTableAlloc()");
      goto Error;
    }

  copy_table->num_files = pe_table->num_pes;
  if (VMMZCopyTableAlloc(copy_table))
    {
      QccErrorAddMessage("(VMMZGetPETable): Error calling VMMZCopyTableAlloc()");
      goto Error;
    }

  for (pe_cnt = 0; pe_cnt < pe_table->num_pes; pe_cnt++)
    {
      if (fscanf(infile, "%d,%x,",
                 &pe_table->pes[pe_cnt].pid,
                 &pe_table->pes[pe_cnt].base_address) != 2)
        {
          QccErrorAddMessage("(VMMZGetPETable): Error reading from vmmz_parse");
          goto Error;
        }
      if (QccFileReadLine(infile, pe_table->pes[pe_cnt].filename))
        {
          QccErrorAddMessage("(VMMZGetPETable): Error reading from vmmz_parse");
          goto Error;
        }
      QccStringCopy(copy_table->files[pe_cnt].filename,
                    pe_table->pes[pe_cnt].filename);
    }
  
  for (process_cnt = 0; process_cnt < process_table->num_processes;
       process_cnt++)
    {
      if (fscanf(infile, "%d,%d\n",
                 &process_table->page_tables[process_cnt].pid,
                 &process_table->page_tables[process_cnt].length) != 2)
        {
          QccErrorAddMessage("(VMMZGetPETable): Error reading from vmmz_parse");
          goto Error;
        }
      
      if (VMMZPageTableAlloc(&process_table->page_tables[process_cnt]))
        {
          QccErrorAddMessage("(VMMZGetPETable): Error calling VMMZPageTableAlloc()");
          goto Error;
        }
      
      for (page_cnt = 0;
           page_cnt < process_table->page_tables[process_cnt].length;
           page_cnt++)
        if (fscanf(infile, "%x,%x,%x\n",
                   &process_table->page_tables[process_cnt].page_table[page_cnt].virtual_address,
                   &process_table->page_tables[process_cnt].page_table[page_cnt].physical_address,
                   &process_table->page_tables[process_cnt].page_table[page_cnt].page_size) != 3)
          {
            QccErrorAddMessage("(VMMZGetPETable): Error reading from vmmz_parse");
            goto Error;
          }
    }
         
  return_value = 0;
  goto Return;
 Error:
  return_value = 1;
  VMMZUnmountVMDK(pe_table->mount_point);
 Return:
  if (infile != NULL)
    fclose(infile);
  return(return_value);
}


static int VMMZInsertPE(unsigned char *memory,
                        long memory_size,
                        const VMMZPEEntry *pe,
                        const VMMZProcessTable *process_table,
                        VMMZCopyFile *copy_file,
                        const QccString mount_point)
{
  int return_value = 0;
  VMMZPEImage pe_image;
  int process_cnt = 0;
  VMMZPageTable *page_table = NULL;
  QccString fullpath;
  
  VMMZPEImageInitialize(&pe_image);
  
  if (process_table == NULL)
    return(0);
  if (pe == NULL)
    return(0);
  
  QccStringSprintf(fullpath, "%s/%s",
                   mount_point,
                   pe->filename);
  
  if (VMMZPEImageRead(fullpath,
                      &pe_image))
    {
      QccErrorAddMessage("(VMMZInsertPE): Error calling VMMZPEImageRead()");
      goto Error;
    }
  
  for (process_cnt = 0; process_cnt < process_table->num_processes;
       process_cnt++)
    if (process_table->page_tables[process_cnt].pid == pe->pid)
      page_table = &process_table->page_tables[process_cnt];
  
  if (page_table == NULL)
    {
      QccErrorAddMessage("(VMMZInsertPE): Error cannot find PID in page table");
      goto Error;
    }
  
  if (VMMZPEImageToMemory(&pe_image,
                          pe->base_address,
                          memory,
                          memory_size,
                          page_table,
                          copy_file))
    {
      QccErrorAddMessage("(VMMZInsertPE): Error calling VMMZPEImageToMemory()");
      goto Error;
    }
  
  return_value = 0;
  goto Return;
 Error:
  return_value = 1;
 Return:
  VMMZPEImageFree(&pe_image);
  return(return_value);
}


int VMMZInsertPEs(unsigned char *memory,
                  long memory_size,
                  const VMMZPETable *pe_table,
                  const VMMZProcessTable *process_table,
                  VMMZCopyTable *copy_table,
                  const QccString predicted_vmem_filename)
{
  int return_value = 0;
  int pe_cnt;
  FILE *outfile = NULL;
  
  if (pe_table == NULL)
    return(0);
  if (memory == NULL)
    return(0);

  for (pe_cnt = 0; pe_cnt < pe_table->num_pes; pe_cnt++)
    if (VMMZInsertPE(memory,
                     memory_size,
                     &pe_table->pes[pe_cnt],
                     process_table,
                     &copy_table->files[pe_cnt],
                     pe_table->mount_point))
      {
        QccErrorAddMessage("(VMMZInsertPEs): Error calling VMMZInsertPE()");
        goto Error;
      }
  
  if (predicted_vmem_filename != NULL)
    {
      if ((outfile = QccFileOpen(predicted_vmem_filename, "w")) == NULL)
        {
          QccErrorAddMessage("(VMMZInsertPEs): Error calling QccFileOpen()");
          goto Error;
        }
      
      if (fwrite((void *)memory,
                 sizeof(unsigned char),
                 memory_size,
                 outfile) != memory_size)
        {
          QccErrorAddMessage("(VMMZInsertPEs): Error writing file");
          goto Error;
        }
    }
  
  return_value = 0;
  goto Return;
 Error:
  return_value = 1;
 Return:
  QccFileClose(outfile);
  return(return_value);
}
