/*
 *
 * VMMZ: Delta Encoding of Virtual-Machine Memory
 * Copyright (C) 2015-2016  James E. Fowler
 *
 * This file is a part of VMMZ.
 *
 * VMMZ is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with VMMZ; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


#include "vmmz.h"


void VMMZPEImageInitialize(VMMZPEImage *pe_image)
{
  if (pe_image != NULL)
    {
      pe_image->size = 0;
      pe_image->image = NULL;
    }
}

int VMMZPEImageAlloc(VMMZPEImage *pe_image)
{
  if (pe_image != NULL)
    if (pe_image->size > 0)
      if ((pe_image->image =
           (unsigned char *)malloc(sizeof(unsigned char) *
                                   pe_image->size)) == NULL)
        {
          QccErrorAddMessage("(VMMZPEImageAlloc): Error allocating memory");
          return(1);
        }

  return(0);
}


void VMMZPEImageFree(VMMZPEImage *pe_image)
{
  if (pe_image != NULL)
    if (pe_image->image != NULL)
      free(pe_image->image);
}


int VMMZPEImageRead(const QccString filename,
                    VMMZPEImage *pe_image)
{
  int return_value = 0;
  FILE *infile = NULL;

  if (pe_image == NULL)
    return(0);
  
  if (VMMZGetFileSize(filename, &pe_image->size))
    {
      QccErrorAddMessage("(VMMZPEImage): Error calling QccFileGetSize()");
      goto Error;
    }
  
  if (VMMZPEImageAlloc(pe_image))
    {
      QccErrorAddMessage("(VMMZPEImage): Error calling VMMZPEImageAlloc()");
      goto Error;
    }
  
  if ((infile = QccFileOpen(filename, "r")) == NULL)
    {
      QccErrorAddMessage("(VMMZPEImage): Error calling QccFileOpen()");
      goto Error;
    }
  
  if ((fread((void *)pe_image->image,
             sizeof(unsigned char),
             (size_t)pe_image->size,
             infile)) != pe_image->size)
    {
      QccErrorAddMessage("(VMMZPEImage): Error reading file");
      goto Error;
    }
  
  return_value = 0;
  goto Return;
 Error:
  VMMZPEImageFree(pe_image);
  return_value = 1;
 Return:
  QccFileClose(infile);
  return(return_value);
}


static int VMMZPEImageHeaderParse(const VMMZPEImage *pe_image,
                                   unsigned int *header_length,
                                   int *num_sections,
                                   VMMZSection **sections)
{
  int return_value = 0;
  int PE_offset;
  int machine_type;
  int optional_header_size;
  int offset_section_headers;
  int section;
  
  PE_offset = GET_UINT32(pe_image->image[OFFSET_PE_SIG]);
  machine_type =
    GET_UINT16(pe_image->image[PE_offset + OFFSET_MACHINE]);
  *num_sections =
    GET_UINT16(pe_image->image[PE_offset + OFFSET_NUMSECTIONS]);
  
  if (machine_type != IMAGE_FILE_MACHINE_I386)
    {
      QccErrorAddMessage("(VMMZPEImageHeaderParse): Unsupported machine type");
      goto Error;
    }
  
  optional_header_size =
    GET_UINT16(pe_image->image[PE_offset + OFFSET_COFF_SIZEOFOPTIONALHEADER]);
  
  *header_length =
    GET_UINT32(pe_image->image[PE_offset + OFFSET_SIZEOFHEADERS]);
                    
  offset_section_headers =
    PE_offset + OFFSET_OPTIONAL_HEADER + optional_header_size;
  
  if ((*sections =
       (VMMZSection *)malloc(sizeof(VMMZSection) * (*num_sections))) == NULL)
    {
      QccErrorAddMessage("(VMMZPEImageHeaderParse): Error allocating memory");
      goto Error;
    }
  
  for (section = 0; section < *num_sections; section++)
    {
      strncpy((*sections)[section].name,
              (char *)(&pe_image->image[offset_section_headers +
                                        SECTION_SIZE * section]), 8);
      (*sections)[section].name[8] = '\0';
      (*sections)[section].virtual_size =
        GET_UINT32(pe_image->image[offset_section_headers +
                                   SECTION_SIZE * section +
                                   OFFSET_SECTION_VIRTUALSIZE]);
      (*sections)[section].virtual_address =
        GET_UINT32(pe_image->image[offset_section_headers +
                                   SECTION_SIZE * section +
                                   OFFSET_SECTION_VIRTUALADDRESS]);
      (*sections)[section].raw_size =
        GET_UINT32(pe_image->image[offset_section_headers +
                                   SECTION_SIZE * section +
                                   OFFSET_SECTION_RAWSIZE]);
      (*sections)[section].raw_ptr =
        GET_UINT32(pe_image->image[offset_section_headers +
                                   SECTION_SIZE * section +
                                   OFFSET_SECTION_RAWPTR]);
    }
  
  return_value = 0;
  goto Return;
 Error:
  return_value = 1;
 Return:
  return(return_value);
}


int VMMZGetPhysicalAddress(unsigned int virtual_address,
                           unsigned int *physical_address,
                           unsigned int *page_size,
                           unsigned int *page_start,
                           const VMMZPageTable *process)
{
  int index = 0;

  if (physical_address != NULL)
    *physical_address = 0;
  if (page_start != NULL)
    *page_start = 0;

  for (index = 0; index < process->length; index++)
    {
      if ((process->page_table[index].virtual_address <=
           virtual_address) &&
          (virtual_address < process->page_table[index].virtual_address +
           process->page_table[index].page_size))
        {
          if (physical_address != NULL)
            *physical_address =
              process->page_table[index].physical_address +
              (virtual_address - process->page_table[index].virtual_address);
          if (page_size != NULL)
            *page_size = process->page_table[index].page_size;
          if (page_start != NULL)
            *page_start = process->page_table[index].physical_address;
          return(0);
        }
      if (process->page_table[index].virtual_address >
          virtual_address)
        {
          if (page_size != NULL)
            *page_size =
              process->page_table[index].virtual_address - virtual_address;
          return(1);
        }
    }
  
  if (page_size != NULL)
    *page_size = (0xffffffff - virtual_address) + 1;

  return(1);
}


static int VMMZCopyToMemory(const VMMZPEImage *pe_image,
                            unsigned int start_address,
                            unsigned int virtual_address,
                            unsigned int length,
                            unsigned char *memory,
                            long memory_size,
                            const VMMZPageTable *process,
                            VMMZCopyFile *copy_file)
{
  int return_value = 0;
  unsigned int physical_address;
  unsigned int page_size;
  unsigned int page_start;
  unsigned int copy_len;
  unsigned int length_to_page_end;
  int length2 = length;
  VMMZCopyEntry copy;
  
  while (length2 > 0)
    {
      if (!VMMZGetPhysicalAddress(virtual_address,
                                  &physical_address,
                                  &page_size,
                                  &page_start,
                                  process))
        {
          length_to_page_end = page_start + page_size - physical_address;
          copy_len =
            (length2 < length_to_page_end) ? length2 : length_to_page_end;
          
          if ((start_address > pe_image->size) ||
              (start_address + copy_len > pe_image->size) ||
              (physical_address > memory_size) ||
              (physical_address + copy_len > memory_size))
            copy_len = (length2 < page_size) ? length2 : page_size;
          else
            {
              memcpy((void *)&memory[physical_address],
                     (void *)&pe_image->image[start_address],
                     (size_t)copy_len);
              
              copy.source = start_address;
              copy.destination = physical_address;
              copy.length = copy_len;
              if (VMMZCopyFileAddCopy(copy_file, &copy))
                {
                  QccErrorAddMessage("(VMMZCopyToMemory): Error calling VMMZCopyFileAddCopy()");
                  goto Error;
                }
            }
        }
      else
        copy_len = (length2 < page_size) ? length2 : page_size;
      
      length2 -= copy_len;
      start_address += copy_len;
      virtual_address += copy_len;
    }
  
  return_value = 0;
  goto Return;
 Error:
  return_value = 1;
 Return:
  return(return_value);
}


int VMMZPEImageToMemory(const VMMZPEImage *pe_image,
                        unsigned int base_address,
                        unsigned char *memory,
                        long memory_size,
                        const VMMZPageTable *page_table,
                        VMMZCopyFile *copy_file)
{
  int return_value = 0;
  unsigned int header_length = 0;
  int num_sections;
  VMMZSection *sections = NULL;
  int section;
  
  if (VMMZPEImageHeaderParse(pe_image,
                             &header_length,
                             &num_sections,
                             &sections))
    {
      QccErrorAddMessage("(VMMZPEImageToMemory): Error calling VMMZPEImageHeaderLength()");
      goto Error;
    }
  
  if (VMMZCopyToMemory(pe_image,
                       0,
                       base_address,
                       header_length,
                       memory,
                       memory_size,
                       page_table,
                       copy_file))
    {
      QccErrorAddMessage("(VMMZPEImageToMemory): Error calling VMMZCopyToMemory()");
      goto Error;
    }
  
  for (section = 0; section < num_sections; section++)
    if (VMMZCopyToMemory(pe_image,
                         sections[section].raw_ptr,
                         base_address + sections[section].virtual_address,
                         sections[section].raw_size,
                         memory,
                         memory_size,
                         page_table,
                         copy_file))
      {
        QccErrorAddMessage("(VMMZPEImageToMemory): Error calling VMMZCopyToMemory()");
        goto Error;
      }
  
  return_value = 0;
  goto Return;
 Error:
  return_value = 1;
 Return:
  if (sections != NULL)
    free(sections);
  return(return_value);
}


